#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define LEFT_CHILD(i)   (((i) * 2) + 1)
#define RIGHT_CHILD(i)  (((i) * 2) + 2)
#define PARENT(i)       (((i) - 1) / 2)
#define INIT_SIZE       2
#define GROW_SIZE	32

struct prioq_head{
	int index;
};

typedef int (*prioq_compare_t)(struct prioq_head *a, struct prioq_head *b);

struct prioq{
	struct prioq_head **heap;	/* heap queue */
	size_t size;			/* size of queue */
	size_t used;			/* number of used positions */
	prioq_compare_t compare;	/* user compare function */
};

static int
ptr_compare(struct prioq_head *a, struct prioq_head *b)
{
	if ((unsigned long) a > (unsigned long) b)
		return 1;
	else if ((unsigned long) a < (unsigned long) b)
		return -1;
	return 0;
}

struct prioq *
prioq_new(prioq_compare_t compare)
{
	struct prioq *queue;

	queue = malloc(sizeof(struct prioq));

	if (queue == NULL) {
		fprintf(stderr, "prioq_new: can't allocate queue\n");
		return NULL;
	}

	memset(queue, 0, sizeof(*queue));

	queue->heap = malloc(sizeof(struct prioq_head *) * INIT_SIZE);

	if (queue->heap == NULL) {
		fprintf(stderr, "prioq_new: can't allocate heap\n");
		free(queue);
		return NULL;
	}

	memset(queue->heap, 0, sizeof(struct prioq_head *) * INIT_SIZE);

	queue->size = INIT_SIZE;
	queue->used = 0;
	queue->compare = ((compare != NULL) ? compare : ptr_compare);

	return queue;
}

#define NEW_SIZE(q)	(((q)->size + GROW_SIZE) * sizeof(struct prioq_head *))
#define OLD_SIZE(q)	((q)->size * sizeof(struct prioq_head *))

int
prioq_put(struct prioq *queue, struct prioq_head *hp)
{
	int i;
	void *tmp;

	if (queue == NULL || hp == NULL)
		fprintf(stderr, "prioq_put: NULL args\n");

	if (queue->size == queue->used) {
		tmp = realloc(queue->heap, NEW_SIZE(queue));
		if (!tmp) {
			fprintf(stderr, "prioq_put: can't realloc heap\n");
			return -1;
		}
		memset(queue->heap + OLD_SIZE(queue), 0,
		       NEW_SIZE(queue) - OLD_SIZE(queue));
		queue->size += GROW_SIZE;
		queue->heap = tmp;
	}

	i = queue->used++;

	while (i && queue->compare(queue->heap[PARENT(i)], hp) < 0) {
 		queue->heap[i] = queue->heap[PARENT(i)];
		queue->heap[i]->index = i;
		i = PARENT(i);
	}

	queue->heap[i] = hp;
	hp->index = i;

	return 0;
}

/*-----------------------------------------------------------*/

struct user_data {
	struct prioq_head head;
	int data;
};

int
compare(struct prioq_head *a, struct prioq_head *b)
{
	struct user_data *ap = (struct user_data *) a;
	struct user_data *bp = (struct user_data *) b;

	if (ap->data > bp->data)
		return 1;
	else if (ap->data < bp->data)
		return -1;
	return 0;
}

int
main(void)
{
	int i;
	struct prioq *q;
	struct user_data a, b, c;

	a.data = 0;
	b.data = 5;
	c.data = 1;

	q = prioq_new(compare);

	prioq_put(q, (struct prioq_head *) &a);
	prioq_put(q, (struct prioq_head *) &b);
	prioq_put(q, (struct prioq_head *) &c);

	for (i = 0; i < q->used; i++) {
		printf("index = %d data = %d\n", i,
		       ((struct user_data *)q->heap[i])->data);
	}

	return 0;
}